# -*- coding: utf-8 -*-
"""Function for auto-generating all physical units from sipy config

`sipy.config.sipy.toml` defines all units which can be retrieved
using `sipy.config.quantity_info`. This module provides a method to
extract the units and create corresponding unit objects.
"""
# Standard Library
from typing import TYPE_CHECKING, Dict, Type

# SIpy
from sipy.config import quantity_info

if TYPE_CHECKING:
    from sipy.quantities.base_quantity import Quantity  # noqa


def generate_units_from_quantities(
    all_quantities: Dict[str, Type["Quantity"]],
) -> Dict[str, "Quantity"]:
    """Generate a dictionary mapping unit names to their corresponding object"""
    my_units = {}
    for quantity in quantity_info():
        for unit_name in quantity.unit_names:
            new_quantity = all_quantities[quantity.name]()
            new_quantity.value = 1
            my_units[unit_name] = new_quantity

        base_unit = new_quantity
        for extra_unit_name in quantity.extra_units:
            new_quantity = all_quantities[quantity.name]()
            new_quantity.value = 1 / getattr(base_unit, extra_unit_name)
            my_units[extra_unit_name] = new_quantity
    return my_units



# -*- coding: utf-8 -*-
"""The API for working with units"""


# SIpy
# flake8: noqa
from sipy.units import prefixes
from sipy.units.base_unit import DerivedSIUnit
from sipy.units.defined_units import generate_units_from_quantities

PUBLIC_UNITS = {"Multiplier": prefixes.UnitMultiple}
PUBLIC_PREFIXES = {
    "Prefix": prefixes.PREFIXES,
    "MassPrefix": prefixes.MASS_PREFIXES,
    "NoPrefix": [],
}



# -*- coding: utf-8 -*-
"""The API for working with quantities"""

# SIpy
# flake8: noqa
from sipy.quantities.base_quantity import quantity_factory
from sipy.quantities.constants import generate_constants
from sipy.quantities.defined_quantities import defined_quantities



# -*- coding: utf-8 -*-
# Standard Library
from typing import Type

# SIpy
from sipy.config import prefixes
from sipy.units.base_unit import SIBaseUnit


class UnitMultiple(SIBaseUnit):
    MULTIPLIER: float = 0

    def _to(self, value: float) -> float:
        return value * self.MULTIPLIER

    def _from(self, value: float) -> float:
        return value / self.MULTIPLIER

    @classmethod
    def factory(cls, name: str, *args):
        new_prefix: Type[UnitMultiple] = type(name, (cls,), {})
        new_prefix.MULTIPLIER = float(args[0])
        return new_prefix


class Prefix(UnitMultiple):
    @classmethod
    def name(cls, unit_name: str = "") -> str:
        return f"{cls.__name__}{unit_name}".lower()


PREFIXES = [Prefix.factory(name, multiplier) for name, multiplier in prefixes().items()]


# Kilograms are problematic for prefixes and need a few tweaks
class MassPrefix(Prefix):
    ADJUSTMENT = 1e-3

    def _to(self, value: float) -> float:
        return value * self.MULTIPLIER * self.ADJUSTMENT

    def _from(self, value: float) -> float:
        return value / self.MULTIPLIER / self.ADJUSTMENT

    @classmethod
    def name(cls, _="") -> str:
        return f"{cls.__name__}grams".lower()


MASS_PREFIXES = [
    MassPrefix.factory(name, multiplier)
    for name, multiplier in prefixes().items()
    if name != "kilo"
]



# -*- coding: utf-8 -*-
"""The base class for an SI Unit

The `sipy.units.si.SIBaseUnit` implements the descriptor protocol.
When combined with a `sipy.quantities.quantity.Quantity` it is used
as an abstraction to a specific unit of that quantity. Every
`sipy.quantities.quantity.Quantity` has a `value` attribute that
stores the underlying value and the `sipy.units.si.SIBaseUnit` acts
as a wrapper around that.
>>> from sipy.quantities.base_quantity import Quantity
>>> q = Quantity
>>> q.unit = SIBaseUnit()
>>> quantity = q()
>>> quantity.unit = 1
>>> quantity.value
1
"""
# Standard Library
from typing import Type


class SIBaseUnit:
    def __get__(self, instance, _) -> float:
        """Implement the descriptor protocol"""
        return self._from(instance.value)

    def __set__(self, instance, value: float):
        """Implement the descriptor protocol"""
        instance.value = self._to(value)

    def _to(self, value: float) -> float:  # pylint: disable=R0201
        """Method that converts to the base value

        Expect this method to be overwritten by the base classes children
        """
        return value

    def _from(self, value: float) -> float:  # pylint: disable=R0201
        """Method that converts from the base value

        Expect this method to be overwritten by the base classes children
        """
        return value

    @classmethod
    def name(cls, unit_name: str = "") -> str:
        """The name of the unit

        It's expected that the name will be used as an attribute to access
        the unit on the `sipy.quantities.quantity.Quantity`

        Args:
            unit_name: The name of the fundamental unit, provided so that child
                classes can generate a modified name e.g. meter -> kilometer

        Returns: The modified name of the unit
        """
        return unit_name.lower()

    @classmethod
    def factory(cls, name: str, *args) -> Type["SIBaseUnit"]:
        """A factory method for generating new types of `sipy.units.si.SIBaseUnit`

        Args:
            name: The name of new `sipy.units.si.SIBaseUnit`
            *args: additional args used by child classes

        Returns:

        """
        raise NotImplementedError()


class DerivedSIUnit(SIBaseUnit):
    """A concrete implementation of the `sipy.units.si.SIBaseUnit` for SIUnits"""

    @classmethod
    def factory(cls, name: str, *args):
        return cls



# -*- coding: utf-8 -*-
# Standard Library
import os
import pathlib
from enum import Enum
from typing import Dict

# Third Party
import toml

# SIpy
from sipy.definitions import unit_counter_from_string

CONFIG_FILE = pathlib.Path(__file__).parent / "sipy.toml"
KEYS = Enum(
    "Keys",
    {
        "prefixes": "Prefixes",
        "quantity": "Quantity",
        "sipy": "Sipy",
        "constants": "Constants",
    },
)
CONFIG_KEYS = Enum(
    "QuantityKey",
    {"si_units": "si_units", "unit_names": "unit_names", "prefix": "prefix"},
)
CONFIG_KEY = "config"
CUSTOM_FILE_ENV_VAR = "SIPY_CFG_FILE"


class SipyConfigError(Exception):
    pass


class Quantity:
    def __init__(self, name, config, extra_units):
        self.name = name
        self.extra_units = extra_units

        self.unit_names = config.get(CONFIG_KEYS.unit_names.value, [])
        self.prefixes = config.get(CONFIG_KEYS.prefix.value, "Prefix")

        if CONFIG_KEYS.si_units.value not in config:
            raise SipyConfigError(
                "Quantities must define 'si_units' in [Quantity.*.config]"
            )
        self.si_units = config[CONFIG_KEYS.si_units.value]

        if self.prefixes and not self.unit_names:
            raise SipyConfigError(
                f"Must specify 'unit_names' for {self.name} if prefix=true"
            )

    @property
    def unit_counter(self):
        return unit_counter_from_string(self.si_units)

    @classmethod
    def from_toml_dict(cls, name, toml_dict):
        config = toml_dict.pop(CONFIG_KEY)
        return cls(name, config, toml_dict)


def _load_config(file_name=CONFIG_FILE):
    cfg = toml.load(file_name)
    validate_config(cfg)
    return cfg


def validate_config(cfg: Dict):
    if KEYS.sipy.value not in cfg:
        raise SipyConfigError(f"[{KEYS.sipy.value}] must be present in config file")

    version = cfg[KEYS.sipy.value].get("version")
    if version != "1.0.0":
        raise SipyConfigError(f"Unsupported config file version: {version}")

    all_keys = [k.value for k in KEYS]
    for key in cfg.keys():
        if key not in all_keys:
            raise SipyConfigError(f"[{key}] is not a valid section name")


def prefixes():
    app_config = _load_config().get(KEYS.prefixes.value, {})

    env_file = os.getenv(CUSTOM_FILE_ENV_VAR)
    if env_file:
        env_config = _load_config(env_file).get(KEYS.prefixes.value, {})
        app_config.update(env_config)

    return app_config


def _extract_quantities(quantities_dict):
    return [
        Quantity.from_toml_dict(name, toml_dict)
        for name, toml_dict in quantities_dict.items()
    ]


def quantity_info():
    app_config = _load_config()[KEYS.quantity.value]
    my_quantities = _extract_quantities(app_config)

    env_file = os.getenv(CUSTOM_FILE_ENV_VAR)
    if env_file:
        env_config = _load_config(env_file).get(KEYS.quantity.value, {})
        my_quantities += _extract_quantities(env_config)

    return my_quantities


def constants():
    app_config = _load_config()[KEYS.constants.value]

    env_file = os.getenv(CUSTOM_FILE_ENV_VAR)
    if env_file:
        env_config = _load_config(env_file).get(KEYS.constants.value, {})
        app_config.update(env_config)

    return app_config



# -*- coding: utf-8 -*-
"""Function for auto-generating all physical constants from sipy config

`sipy.config.sipy.toml` defines all physical constants which can be retrieved
using `sipy.config.constants`. This module provides a method to
extract the constants and create the constants.
"""
# Standard Library
from typing import TYPE_CHECKING, Dict

# SIpy
from sipy.config import constants
from sipy.definitions import unit_counter_from_string
from sipy.quantities.base_quantity import quantity_factory

if TYPE_CHECKING:
    from sipy.quantities.base_quantity import Quantity  # noqa


def generate_constants() -> Dict[str, "Quantity"]:
    """Create physical quantities representing the defined constants"""
    constant_dict = {}
    for constant_name, [value, unit_string] in constants().items():
        quantity = quantity_factory(unit_counter_from_string(unit_string))()
        quantity.value = float(value)
        constant_dict[constant_name] = quantity
    return constant_dict



# -*- coding: utf-8 -*-
"""Function for auto-generating all physical quantities from sipy config

`sipy.config.sipy.toml` defines all quantities which can be retrieved
using `sipy.config.quantity_info`. This module provides a method to
extract the quantities and their create corresponding types.
"""
# Standard Library
from typing import TYPE_CHECKING, Dict, Type

# SIpy
from sipy.config import quantity_info
from sipy.quantities.base_quantity import quantity_factory

if TYPE_CHECKING:
    from sipy.quantities.base_quantity import Quantity  # noqa


def defined_quantities() -> Dict[str, Type["Quantity"]]:
    """Generate a dictionary mapping quantity names to their object type"""
    my_quantities = {}
    for quantity in quantity_info():
        my_quantities[quantity.name] = quantity_factory(quantity.unit_counter)
    return my_quantities



# -*- coding: utf-8 -*-
"""
`sipy` is a python package for manipulating physical quantities. All quantities in `sipy` are
constructed dynamically from their [SI base units](https://en.wikipedia.org/wiki/SI_base_unit) so
that `sipy` can use [dimensional analysis](https://en.wikipedia.org/wiki/Dimensional_analysis) to
spot errors in your calculations and learn all the valid units for a quantity.
>>> from sipy import meter, second
>>> meter
<Length 1.00E+00m>
>>> try:
...     meter + second
... except TypeError as exception:
...     print(exception)
unsupported operand type(s) for +: 'Length' and 'Time'
>>> distance = 1000 * meter
>>> distance.kilometers
1.0

It is also able to handle non-SI base units. The example below demonstrates the various
ways you can create and combine units by calculating the power, in watts, required for a car
to accelerate from 0 mph to 100 mph, in 5 seconds we use

>>> from sipy import Length, Time
>>> distance_to_travel = Length()
>>> distance_to_travel.miles = 100
>>> speed_of_car = distance_to_travel / Time(hours=1)
>>> mass_of_car = 1300 * kilogram
>>> power = 0.5 * mass_of_car * speed_of_car ** 2 / Time(seconds=5)
>>> round(power.watts)
259797

.. NOTE::
    To calculate the power required we determine the change in kinetic energy and divide
    it by the time taken, giving us this equation:
    .. math::
        \\displaystyle \\frac{ \\frac{1}{2} \\times mass \\times {velocity}^2} {time}


As `sipy` dynamically creates all it's quantities it's easy to add new units
todo link to a guide

As well as quantities, it comes packaged with a variety of physical constants
>>> from sipy import avogradro, boltzmann, planck
>>> avogradro
<InverseAmount 6.02E+23mol^-1>
>>> boltzmann
<InverseTemperatureMassLength2InverseTime2 1.38E-23K^-1kgm^2s^-2>
>>> planck
<MassLength2InverseTime 6.63E-34kgm^2s^-1>

todo
    * tune UNITS for perf
"""
# SIpy
from sipy.quantities import defined_quantities, generate_constants
from sipy.units import generate_units_from_quantities

__version__ = "0.3.0"


# Update project namespace to include all quantities, units and constants
_QUANTITIES = defined_quantities()
locals().update(_QUANTITIES)
locals().update(generate_units_from_quantities(_QUANTITIES))
locals().update(generate_constants())



# -*- coding: utf-8 -*-
# Standard Library
import typing
from collections import Counter

# SI base units
SECONDS = "s"
METERS = "m"
KILOGRAMS = "kg"
AMPERES = "A"
KELVINS = "K"
MOLES = "mol"
CANDELAS = "cd"

# SI base quantities
TIME = "Time"
LENGTH = "Length"
MASS = "Mass"
CURRENT = "Current"
TEMPERATURE = "Temperature"
AMOUNT = "Amount"
LUMINOSITY = "Luminosity"

# Mapping SI base units to SI base qunatities
UNIT_QUANTITY = {
    SECONDS: TIME,
    METERS: LENGTH,
    KILOGRAMS: MASS,
    AMPERES: CURRENT,
    KELVINS: TEMPERATURE,
    MOLES: AMOUNT,
    CANDELAS: LUMINOSITY,
}

# SI base units mapped to their full names
UNIT_NAME = {
    SECONDS: "seconds",
    METERS: "meters",
    KILOGRAMS: "kilograms",
    AMPERES: "amperes",
    KELVINS: "kelvins",
    MOLES: "moles",
    CANDELAS: "candelas",
}

SCALAR = "Scalar"
INVERSE = "Inverse"


def unit_counter(  # pylint: disable=R0913
    seconds: int = 0,
    meters: int = 0,
    kilograms: int = 0,
    amperes: int = 0,
    kelvins: int = 0,
    moles: int = 0,
    candelas: int = 0,
) -> typing.Counter[str]:
    return Counter(
        {
            SECONDS: seconds,
            METERS: meters,
            KILOGRAMS: kilograms,
            AMPERES: amperes,
            KELVINS: kelvins,
            MOLES: moles,
            CANDELAS: candelas,
        }
    )


def unit_string(
    units: typing.Union[typing.Counter[str], typing.Dict[str, float]]
) -> str:
    """Construct a string from a units counter

    :param units: The unit count
    :return: A string representing the units

    >>> unit_string(Counter({"a": 1, "b": 0, "c": 0}))
    'a'
    >>> unit_string(Counter({"a": 0, "b": 1, "c": 0}))
    'b'
    >>> unit_string(Counter({"a": 2, "b": 0, "c": 0}))
    'a^2'
    >>> unit_string(Counter({"a": -1, "b": 0, "c": 0}))
    'a^-1'
    >>> unit_string(Counter({"a": 1, "b": 2, "c": 0}))
    'ab^2'
    """
    string = ""
    for unit, power in sorted(units.items()):
        if power == 1:
            string += unit
        elif power != 0:
            string += f"{unit}^{power}"
    return string


def unit_counter_from_string(unit_repr: str):
    """

    Args:
        unit_string:

    Returns:

    >>> unit_counter_from_string("m")
    Counter({'m': 1, 's': 0, 'kg': 0, 'A': 0, 'K': 0, 'mol': 0, 'cd': 0})
    >>> unit_counter_from_string("m^2")
    Counter({'m': 2, 's': 0, 'kg': 0, 'A': 0, 'K': 0, 'mol': 0, 'cd': 0})
    >>> unit_counter_from_string("m^-1")
    Counter({'s': 0, 'kg': 0, 'A': 0, 'K': 0, 'mol': 0, 'cd': 0, 'm': -1})
    >>> unit_counter_from_string("m.s")
    Counter({'s': 1, 'm': 1, 'kg': 0, 'A': 0, 'K': 0, 'mol': 0, 'cd': 0})
    >>> unit_counter_from_string("kg.m^2.s^-2")
    Counter({'m': 2, 'kg': 1, 'A': 0, 'K': 0, 'mol': 0, 'cd': 0, 's': -2})
    """
    units = unit_repr.split(".")
    counter_dict = {}
    for unit in sorted(units):
        unit_info = unit.split("^")
        if len(unit_info) == 1:
            count = "1"
            (unit_char,) = unit_info
        else:
            unit_char, count = unit_info
        counter_dict[UNIT_NAME[unit_char]] = int(count)

    return unit_counter(**counter_dict)


def unit_reference(units: typing.Counter[str]) -> str:
    """Construct a name that can be used to reference a unit

    :param units: The unit count
    :return: A string representing the refernce

    >>> unit_reference(unit_counter(seconds=1))
    'seconds'
    >>> unit_reference(unit_counter(meters=2))
    'meters2'
    >>> unit_reference(unit_counter(kilograms=-1))
    'perkilogram'
    >>> unit_reference(unit_counter(meters=1, seconds=-1))
    'meterspersecond'
    >>> unit_reference(unit_counter(meters=-1, seconds=1))
    'secondspermeter'
    >>> unit_reference(unit_counter(seconds=-2, meters=1, kilograms=2))
    'kilograms2meterspersecond2'
    >>> unit_reference(unit_counter(seconds=-2, meters=-1, kilograms=2))
    'kilograms2permeterpersecond2'
    >>> unit_reference(unit_counter())
    Traceback (most recent call last):
    ...
    ValueError: Count cannot be all zeros
    """
    if all(value == 0 for value in units.values()):
        raise ValueError("Count cannot be all zeros")

    start = end = ""
    # Sort based on unit, i.e. A, K, cd, kg, m, mol, s
    for unit, count in sorted(units.items()):
        if count != 0:
            quantity_name = UNIT_NAME[unit]

            if count == 1:
                start += f"{quantity_name}"
            elif count == -1:
                end += f"per{quantity_name[:-1]}"
            elif count > 1:
                start += f"{quantity_name}{count}"
            else:
                end += f"per{quantity_name[:-1]}{-count}"

    return start + end


def quantity_string(units: typing.Counter[str]) -> str:
    """Determine a quantity name from it's units

    :param units: The units of the class
    :return: The name of the class

    >>> quantity_string(unit_counter(seconds=1))
    'Time'
    >>> quantity_string(unit_counter(meters=2))
    'Length2'
    >>> quantity_string(unit_counter(kilograms=-1))
    'InverseMass'
    >>> quantity_string(unit_counter(meters=1, seconds=-1))
    'LengthInverseTime'
    >>> quantity_string(unit_counter(seconds=-2, meters=1, kilograms=2))
    'Mass2LengthInverseTime2'
    """
    name = ""

    # Sort based on unit, i.e. A, K, cd, kg, m, mol, s
    for unit, count in sorted(units.items()):
        if count != 0:
            quantity = UNIT_QUANTITY[unit]
            if count < 0:
                name += INVERSE
                count = -count

            if count == 1:
                name += f"{quantity}"
            else:
                name += f"{quantity}{count}"

    return name if name else SCALAR



# -*- coding: utf-8 -*-
"""Populate the UNIT_MAP dictionary used when generating new quantities.

 The `sipy.quantities.quantity.quantity_factory` looks up new quantities in
 the UNIT_MAP to determine what units they should have.
"""
# Standard Library
from collections import defaultdict
from typing import Counter, DefaultDict, List, Tuple, Type

# SIpy
from sipy.config import quantity_info
from sipy.definitions import unit_reference, unit_string
from sipy.units import PUBLIC_PREFIXES, PUBLIC_UNITS, DerivedSIUnit

# A mapping from SI measures to supported prefixes
UNIT_MAP: DefaultDict[str, List[Tuple[str, Type[DerivedSIUnit]]]] = defaultdict(list)


def add_unit_mapping(
    quantity_units: Counter,
    unit_class: Type[DerivedSIUnit] = DerivedSIUnit,
    name: str = None,
):
    """Register a new type of unit in the UNIT_MAP

    Args:
        quantity_units: The SI Units associated with the unit
        unit_class: The `sipy.units.si.SIBaseUnit` to register
        name: The name of fundamental unit
    """
    name = name or unit_reference(quantity_units)
    UNIT_MAP[unit_string(quantity_units)] += [(name, unit_class)]


def add_unit_mappings(
    quantity_units: Counter, unit_classes: List[Type[DerivedSIUnit]], name: str = None,
):
    """Register multiple units for a single quantity in the UNIT_MAP

    Args:
        quantity_units: The SI Units associated with the unit
        unit_classes: A list of `sipy.units.si.SIBaseUnit` to register
        name: The name of fundamental unit
    """
    for unit_class in unit_classes:
        add_unit_mapping(quantity_units, unit_class, name)


def populate_map():
    """Populate UNIT_MAP with new units by reading from the packates config

    The function can be called multiple times in order to re-populate the
    UNIT_MAP. This will cause us to re-read the packages config, which
    might have changed.
    """
    UNIT_MAP.clear()
    for quantity in quantity_info():
        unit_count = quantity.unit_counter

        # Some units have custom prefixies defined. A custom prefix is specified
        # by the quantity as a string that we lookup in the public dictionary
        # of prefixes.
        prefixes = PUBLIC_PREFIXES[quantity.prefixes]

        # Add the basic unit names (ones that refer to the quantities fundamental
        # unit) and all the prefixes associated with quantity, e.g. Kilo, Mega, ...
        for unit_name in quantity.unit_names:
            add_unit_mapping(unit_count, DerivedSIUnit, unit_name)
            add_unit_mappings(unit_count, prefixes, unit_name)

        # Add in all the custom units, the units type is specified as a string that
        # is looked up in the dictionary of public unit types.
        for extra_unit_name, unit_info in quantity.extra_units.items():
            unit_base_type, *args = unit_info
            unit_type = PUBLIC_UNITS[unit_base_type].factory(extra_unit_name, *args)
            add_unit_mapping(unit_count, unit_type, extra_unit_name)


populate_map()



# -*- coding: utf-8 -*-
# pylint: disable=R0401
"""The base class for physical quantities and associated factory.

This module implements `sipy.quantities.quantity.Quantity`, the base class
for all physical quantities implemented by SIpy. The
`sipy.quantities.quantity.quantity_factory` provides a method of generating
new quantity classes based on their SI units.

Example:
    The `Length` class could be generated with the following function call:

    >>> Length = quantity_factory(Counter(m=1))
    >>> print(Length())
    0.00E+00m
"""
# Standard Library
import operator
import typing
from collections import Counter
from copy import copy
from numbers import Number

# SIpy
from sipy.automap import UNIT_MAP
from sipy.definitions import quantity_string, unit_string


class Quantity:
    """Base class for all physical quantities.

    All new SIpy physical quantities should inherit from
    `sipy.quantities.quantity.Quantity`. Children should override the `UNITS`
    class attribute with a `collections.Counter` indicating the SI units of
    the quantity. The keys for `UNITS` are those defined in `sipy.definitions`.

    `sipy.quantities.quantity.Quantity` implements support for the standard
    mathematical operations that should work between physical quantities. It
    uses dimensional analysis to identify (and police) the operations
    allowed between different quantities as well as calculating the resulting
    quantities.

    Dimensional Analysis Rules:
        * Addition and subtraction is only allowed between quantities that have
        the same physical type. The result of the operation also has that type.
        * Multiplication and division works between quantities of different
        types and may produce a quantity of a different type.
        * Power operations are allowed if they result in a valid physical
        quantity.
        * Two quantities are equal if they have the same value ans are of the
        same type.
        * Quantities can be compared if they are of the same type.

    Attributes:
        UNITS: the SI units of the quantity, overridden by child classes.
        value: the value of the quantity expressed in SI units.
    """

    UNITS: typing.Counter[str] = Counter({})

    def __init__(self, **kwargs: str):
        """Create a quantity with a value of zero

        Args:
            **kwargs: an optional kwarg used to initialise the quantity with
                a non-zero value. The name of kwarg should correspond to
                a unit of measurement supported by the quantity. For instance

                >>> from sipy import Length
                >>> Length(kilometers=2.5)
                <Length 2.50E+03m>
        """
        self.value: float = 0
        if len(kwargs) > 1:
            raise TypeError(
                f"{self.__class__.__name__}.__init__() takes at most 1 keyword argument"
            )
        for unit, value in kwargs.items():
            if not hasattr(self, unit):
                raise AttributeError(
                    f"'{self.__class__.__name__}' has no attribute '{unit}'"
                )
            setattr(self, unit, value)

    def __add__(self, other: "Quantity") -> "Quantity":
        if not self._isinstance(other):
            raise TypeError(
                f"unsupported operand type(s) for +: "
                f"'{self.__class__.__name__}' and '{other.__class__.__name__}'"
            )
        val = self.__class__()
        val.value = self.value + other.value
        return val

    def __sub__(self, other: "Quantity") -> "Quantity":
        if not self._isinstance(other):
            raise TypeError(
                f"unsupported operand type(s) for -: "
                f"'{self.__class__.__name__}' and '{other.__class__.__name__}'"
            )
        val = self.__class__()
        val.value = self.value - other.value
        return val

    def __mul__(self, other: typing.Union["Quantity", float]) -> "Quantity":
        if not isinstance(other, Quantity):
            if not isinstance(other, Number):
                raise TypeError(
                    f"unsupported operand type(s) for *: "
                    f"'{self.__class__.__name__}' and '{other.__class__.__name__}'"
                )
            return self._number_mul(other)

        quantity = quantity_factory(add_counters(self.UNITS, other.UNITS))()
        quantity.value = self.value * other.value
        return quantity

    def __rmul__(self, other: float) -> "Quantity":
        if not isinstance(other, Number):
            raise TypeError(
                f"unsupported operand type(s) for *: "
                f"'{other.__class__.__name__}' and '{self.__class__.__name__}'"
            )

        return self._number_mul(other)

    def __truediv__(self, other: typing.Union["Quantity", float]) -> "Quantity":
        if not isinstance(other, Quantity):
            if not isinstance(other, Number):
                raise TypeError(
                    f"unsupported operand type(s) for /: "
                    f"'{self.__class__.__name__}' and '{other.__class__.__name__}'"
                )
            return self._number_truediv(other)

        quantity = quantity_factory(subtract_counters(self.UNITS, other.UNITS))()
        quantity.value = self.value / other.value
        return quantity

    def __rtruediv__(self, other: float) -> "Quantity":
        if not isinstance(other, Number):
            raise TypeError(
                f"unsupported operand type(s) for /: "
                f"'{other.__class__.__name__}' and '{self.__class__.__name__}'"
            )

        inverted_units = Counter({unit: -power for unit, power in self.UNITS.items()})
        quantity = quantity_factory(inverted_units)()
        quantity.value = other / self.value
        return quantity

    def __pow__(self, power: float) -> "Quantity":
        new_units = multiply_counter(self.UNITS, power)
        quantity = quantity_factory(new_units)()
        quantity.value = self.value ** power
        return quantity

    def __str__(self) -> str:
        return f"{self.value:.2E}{self.unit_string}"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.value:.2E}{self.unit_string}>"

    def __eq__(self, other: object) -> bool:
        if not self._isinstance(other):
            return False
        return self.value == other.value  # type: ignore

    def __lt__(self, other: object) -> bool:
        if not self._isinstance(other):
            raise TypeError(
                f"'<' not supported between instances of "
                f"'{self.__class__.__name__}' and '{other.__class__.__name__}'"
            )
        return self.value < other.value  # type: ignore

    def _isinstance(self, other):
        """Determine if 2 Quantities are of the same type

        >>> from sipy import meters, seconds
        >>> seconds._isinstance(meters)
        False
        >>> seconds._isinstance(seconds)
        True
        """
        return self.__class__.__name__ == other.__class__.__name__

    @property
    def unit_string(self) -> str:
        """The human readable form of this quantities units, e.g. m^2.

        Returns: The qunatities units as a string.
        """
        return unit_string(self.UNITS)

    def _number_mul(self, number: float) -> "Quantity":
        """Scalar multiplication for the quantity."""
        val = self.__class__()
        val.value = self.value * number
        return val

    def _number_truediv(self, number: float) -> "Quantity":
        """Scalar division for the quantity."""
        val = self.__class__()
        val.value = self.value / number
        return val


def quantity_factory(units: Counter) -> typing.Type[Quantity]:
    """Construct a Quantity class based on the provided units

    >>> Q = quantity_factory(Counter({"m": 1}))
    >>> Q.__name__, Q.UNITS
    ('Length', Counter({'m': 1}))
    >>> d = Q()
    >>> d.kilometers = 1
    >>> d.meters, d.kilometers
    (1000.0, 1.0)

    Args:
        units: The dimensions of the quantity

    Returns:
        The corresponding quantity class
    """
    unit_combo = unit_string(units)
    name = quantity_string(units)
    quantity: typing.Type[Quantity] = type(name, (Quantity,), {})
    quantity.UNITS = units

    if unit_combo == "":
        return quantity

    for unit_name, unit in UNIT_MAP[unit_combo]:
        setattr(quantity, unit.name(unit_name), unit())

    return quantity


def add_counters(
    first: typing.Counter[str], second: typing.Counter[str]
) -> typing.Counter[str]:
    """Add together two `collection.Counter`.

    >>> add_counters(Counter(a=-2, b=3, c=4), Counter(a=5, d=1))
    Counter({'c': 4, 'a': 3, 'b': 3, 'd': 1})

    Args:
        first: The initial counter.
        second: The counter to add to the initial counter.

    Returns:
        The `collection.Counter` resulting from the addition
    """
    return _combine_counters(first, second, operator.add)


def subtract_counters(
    first: typing.Counter[str], second: typing.Counter[str]
) -> typing.Counter[str]:
    """Subtract two `collection.Counter`.

    >>> subtract_counters(Counter(a=-2, b=3, c=4), Counter(a=5, d=1))
    Counter({'c': 4, 'b': 3, 'd': -1, 'a': -7})

    Args:
        first: The initial counter.
        b: The counter to subtract from the initial counter.

    Returns:
        The `collection.Counter` resulting from the subtraction
    """
    return _combine_counters(first, second, operator.sub)


def _combine_counters(
    first: typing.Counter[str], second: typing.Counter[str], combiner: typing.Callable
) -> typing.Counter[str]:
    """Combine two counters using the specified operator"""
    tmp = copy(first)
    for item, value in second.items():
        tmp[item] = combiner(tmp[item], value)

    return tmp


def multiply_counter(
    counter: typing.Counter[str], factor: float
) -> typing.Counter[str]:
    """Scale a `collection.Counter` by a factor.

    >>> multiply_counter(Counter(a=-2, b=3), 2)
    Counter({'b': 6, 'a': -4})

    Args:
        counter: The initial counter.
        factor: The factor to scale counter by.

    Returns:
        The scaled `collection.Counter`

    Raises:
        ValueError: The values of the resulting counter
            are not whole numbers.
    """
    tmp_counter: typing.Dict[str, float] = {}
    for unit in counter:
        tmp_counter[unit] = counter[unit] * factor

    if any(v % 1 for v in tmp_counter.values()):
        raise ValueError(
            f"Quantity must have integer units: {unit_string(tmp_counter)}"
        )

    return Counter(tmp_counter)
